#include "Camera.h"

void Camera::Frustum() {
	m_P.set(2.0f * fNear / (fRight - fLeft), 0.0f, (fRight + fLeft) / (fRight - fLeft), 0.0f,
			0.0f, (2.0f * fNear) / (fTop - fBottom), (fTop + fBottom) / (fTop - fBottom), 0.0f,
			0.0f, 0.0f, -(fFar + fNear) / (fFar - fNear), (-2.0f * fFar * fNear) / (fFar - fNear),
			0.0f, 0.0f, -1.0f, 0.0f);
}

void Camera::LookAt() {
	
	gmtl::Vec3f f = m_vecCenter - m_vecPosition;	
	gmtl::normalize(f);

	gmtl::Vec3f s;
	gmtl::cross(s, f, m_vecUp);
	gmtl::normalize(s);

	gmtl::Vec3f u;
	gmtl::cross(u, s, f);

	m_C.set( s[0],  s[1],  s[2], -s[0] * m_vecPosition[0] - s[1] * m_vecPosition[1] - s[2] * m_vecPosition[2],
			 u[0],  u[1],  u[2], -u[0] * m_vecPosition[0] - u[1] * m_vecPosition[1] - u[2] * m_vecPosition[2],
			-f[0], -f[1], -f[2],  f[0] * m_vecPosition[0] + f[1] * m_vecPosition[1] + f[2] * m_vecPosition[2],
			 0.0f,  0.0f,  0.0f, 1.0f);	
}

Camera::Camera() : m_bProjectionInit(false) {
	m_vecPosition = gmtl::Vec3f(0.0f, 0.0f, 0.0f);
	m_vecCenter = gmtl::Vec3f(0.0f, 0.0f, -1.0f);
	m_vecUp = gmtl::Vec3f(0.0f, 1.0f, 0.0f);	
}

Camera::Camera(float l, float r, float b, float t, float n, float f) : m_bProjectionInit(false) {
	Camera();	
	setProjection(l, r, b, t, n, f);	
}

Camera::Camera(float fovy, float aspect, float n, float f) {
	Camera();
	setProjection(fovy, aspect, n, f);
}

void Camera::setProjection(float l, float r, float b, float t, float n, float f) {
	const float fPI = 3.1415926535897932384626433832795f;

	float lFar = (f / n) * l;
	float rFar = (f / n) * r;
	float bFar = (f / n) * b;
	float tFar = (f / n) * t;
	
	m_vtVertices[0] = gmtl::Vec4f(l, b, -n, 1.0f);
	m_vtVertices[1] = gmtl::Vec4f(r, b, -n, 1.0f);
	m_vtVertices[2] = gmtl::Vec4f(r, t, -n, 1.0f);
	m_vtVertices[3] = gmtl::Vec4f(l, t, -n, 1.0f);
	m_vtVertices[4] = gmtl::Vec4f(lFar, bFar, -f, 1.0f);
	m_vtVertices[5] = gmtl::Vec4f(rFar, bFar, -f, 1.0f);
	m_vtVertices[6] = gmtl::Vec4f(rFar, tFar, -f, 1.0f);
	m_vtVertices[7] = gmtl::Vec4f(lFar, tFar, -f, 1.0f);

	m_bProjectionInit = true;

	gmtl::Matrix44f Cinv = m_C;
	gmtl::invert(Cinv);
		
	for(int i = 0; i < 8; i++)
		m_vtVerticesTransformed[i] = Cinv * m_vtVertices[i];		
	
	fLeft = l;
	fRight = r;
	fBottom = b;
	fTop = t;
	fNear = n;
	fFar = f;

	fovy = (360.0f / fPI) * atan(fTop / fNear);

	Frustum();	
}

void Camera::setProjection(float fovy, float aspect, float n, float f) {
	const float fPI = 3.1415926535897932384626433832795f;

	float l, r, b, t;
	t = n * tan(fovy * fPI / 360.0f);
	b = -t;
	l = b * aspect;
	r = t * aspect;
	
	setProjection(l, r, b, t, n, f);
}

void Camera::setLocation(float posx, float posy, float posz, 
			   float centerx, float centery, float centerz,	
			   float upx, float upy, float upz) {
	if(!m_bProjectionInit) {
		std::cerr << "Camera error: The projection has to be specified first." << std::endl;
		exit(-1);
	}
	
	m_vecPosition = gmtl::Vec3f(posx, posy, posz);
	m_vecCenter = gmtl::Vec3f(centerx, centery, centerz);
	m_vecUp = gmtl::Vec3f(upx, upy, upz);

	LookAt();
	
	gmtl::Vec3f vecDir = m_vecCenter - m_vecPosition;
	gmtl::normalize(vecDir);

	m_rayLineOfSight.setOrigin(m_vecPosition);
	m_rayLineOfSight.setDir(vecDir);

	gmtl::Matrix44f Cinv = m_C;
	gmtl::invert(Cinv);
	
	for(int i = 0; i < 8; i++) {
		m_vtVerticesTransformed[i] = Cinv * m_vtVertices[i];		
	}
}

void Camera::setOrbit(float rotx, float roty, float trans) {
	const float PIDIV180 = 3.1415926535897932384626433832795f / 180.0f;

	if(!m_bProjectionInit) {
		std::cerr << "Camera error: The projection has to be specified first." << std::endl;
		exit(-1);
	}
	
	gmtl::Matrix44f T, R1, R2, M;

	gmtl::setTrans(T, gmtl::Vec3f(0.0f, 0.0f, -trans));
	
	float f = roty * PIDIV180;
	
	gmtl::setRot(M, gmtl::AxisAnglef(-f, gmtl::Vec3f(0.0f, 1.0f, 0.0f)));	
	
	gmtl::Vec4f v(1.0f, 0.0f, 0.0f, 1.0f);	
	v = M * v;
	gmtl::Vec3f u(v[0], v[1], v[2]);
	gmtl::normalize(u);

	gmtl::transpose(R1, M);
	
	gmtl::identity(M);
	f = (rotx * 3.1415926535897932384626433832795f) / 180.0f;
	gmtl::setRot(M, gmtl::AxisAnglef(-f, gmtl::Vec3f(u[0], u[1], u[2])));
	gmtl::transpose(R2, M);

	m_C = T * R1 * R2;

	gmtl::Matrix44f	Cinv = m_C;
	
	gmtl::invert(Cinv);
	
	for(int i = 0; i < 8; i++) {
		m_vtVerticesTransformed[i] = Cinv * m_vtVertices[i];
	}

	m_vecPosition[0] = -static_cast<float>(trans * cos(rotx * PIDIV180) * sin(roty * PIDIV180));
    m_vecPosition[1] = static_cast<float>(trans * sin(rotx * PIDIV180));               
	m_vecPosition[2] = static_cast<float>(trans * cos(rotx * PIDIV180) * cos(roty * PIDIV180));
}